/*
  Open a file, then 'exec' to another program.

  USAGE: open path flags command ...

    'path'     is a pathname of the file to be opened
    'flags'    currently takes the value "r", "w" or "rw"
    'command'  is a pathname of the command to be exec'd
    '...'      arguments for the exec'd command (should
	       include at least tail of command pathname).

  The result of the 'open' call is placed in a new environment variable
  'fildes'.

  WARNING: Setting 'command' to 'csh' is useless, because 'csh'
	   appears to explicitly close descriptors on startup.
	   Use 'sh' instead.
*/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
char *calloc () ;

/* Stuff for reporting system errors */

extern int  errno, sys_nerr ;
extern char *sys_errlist [] ;

static char *USAGE = "USAGE: open path r|w|rw command argument ..." ;

/* Options of 'flags' argument */

#define NUM_OPTIONS 3 

static char *options [NUM_OPTIONS] = {"r", "w", "rw"} ;
static  int    masks [NUM_OPTIONS] = {O_RDONLY, O_WRONLY, O_RDWR} ;

/* Max size environment string allowed by 'setenv' */

#define MAX_STR_SIZE 256


int get_flags (str)
char *str ;
{
  /*
  Convert string ("r", "w", "rw", ...) to flag mask for 'open' call
  Returns -ve value if invalid argument.
  */

  int mask  = -1 ;
  int index =  0 ;

  while ((mask < 0) && (index < NUM_OPTIONS))
    if (!strcmp (str, options [index]))
      mask = masks [index] ;
    else
      index++ ;

  return mask ;
}

int open_file (path, flag_str)
char *path, *flag_str ;
{
  /*
    Open file and return a descriptor.
    'flag_str' encodes the flags for the call to 'open'.
    Descriptor is -ve if call fails.
  */

  int flags, fildes ;

  if ((flags = get_flags (flag_str)) >= 0)
  {
    fildes  = open (path, flags) ;
    if (fildes < 0)
    {
      if ((errno > 0) && (errno < sys_nerr))
	fprintf (stderr, "ERROR: %s\n", sys_errlist [errno]) ;
      else
	fprintf (stderr, "ERROR: Failed to open file \"%s\"\n",
			 path) ;
    }
  }
  else
  {
    fprintf (stderr, "ERROR: unrecognised flag \"%s\" for \'open\'\n",
		     flag_str) ;
    fildes  = flags ;
  }
  return fildes ;
}

int name_matches (pair, name)
char *pair, *name ;
{
  /*
    Returns true iff 'pair' has value "str1=str2", where
    the "str1" is the value of 'name'.
  */

  char *index ;

  if (index = strchr (pair, '='))
  {
    int result ;

    *index = 0 ;
    result = !strcmp (pair, name) ;
    *index = '=' ;
    return result ;
  }
  else
    return 0 ;
}

char **setenv(old_envp, name, value)
char **old_envp ;
char  *name, *value ;
{
  /*
    Return a new environment, modified by inclusion of 'name'='value'
    pair.  If 'name' is already in the environment, overwrite the
    old assignment.
  */

  char **new_envp ;

  int new_loc ;
  int exists  ;

  {
    /*
      Check whether name already exists in environment, and
      find location name-value pair will have in new environment.
    */

    int found = 0 ;

    new_loc = 0 ;
    exists  = 0 ;
    while (!found)
    {
      char* pair = old_envp [new_loc] ;

      if (pair)
      {
	if (name_matches (pair, name))
	{
	  found  = 1 ;
	  exists = 1 ;
	}
	else
	  new_loc++ ;
      }
      else
	found = 1 ;
    }
  }

  /*
    If new environment is bigger than old, malloc space for the new
    vector, and copy the pointers in the old environment vector
    to it.
  */

  if (exists)
    new_envp = old_envp ;
  else
  {
    int i ;

    new_envp = (char **) calloc (new_loc + 2, sizeof (char *)) ;
    for (i = 0; i < new_loc; i++)
      new_envp [i] = old_envp [i] ;
  }

  /* Malloc space for the new pair, and copy value to it */

  new_envp [new_loc] = calloc (MAX_STR_SIZE, sizeof (char)) ;
  sprintf (new_envp [new_loc], "%s=%s", name, value) ;

  return new_envp ;
}

void main (argc, argv, envp)
int   argc   ;
char *argv[] ;
char *envp[] ;
{
  char **new_envp ;
  int fildes ;
  char str [64] ;

  if (argc < 5)
  {
    fprintf (stderr, "%s\n", USAGE) ;
    exit (1) ;
  }

  fildes = open_file (argv [1], argv [2]) ;
  sprintf (str, "%d", fildes) ;
  new_envp = setenv (envp, "fildes", str) ;

  /*
  {
    int new_envc = 0;
    while (new_envp [new_envc])
    {
      printf ("%s\n", new_envp [new_envc]) ;
      new_envc++ ;
    }
  }
  */

  execve (argv [3], argv + 4, new_envp) ;
}

